{$A+,B-,D+,E-,F-,G+,I-,L+,N+,O-,P-,Q-,R-,S-,T-,V+,X+,Y+}
{by Behdad Esfahbod}
{Define Objects}
unit
  _Expression;

interface

  uses
    Objects, _Global;

  type
    PIdentifier = ^ TIdentifier;
    TIdentifier = object (TObject)
      Name : array [0 .. MaxIdentifier] of TIdent;
      Size : Integer;
      constructor Init;
      function  Insert (S : TIdent) : Integer; virtual;
      function  Index  (S : TIdent) : Integer;
      destructor Done; virtual;
    end;

  type
    PNumber = ^ TNumber;
    TNumber = object (TIdentifier)
      function  Insert (S : TIdent) : Integer; virtual;
      function  Index  (S : TIdent) : Integer;
      function  InsertNum (X : Extended) : Integer; virtual;
      function  IndexNum  (X : Extended) : Integer;
    end;

  type
    PVariable = ^ TVariable;
    TVariable = object (TIdentifier)
      function  Insert (S : TIdent) : Integer; virtual;
    end;

  type
    POperator = ^ TOperator;
    TOperator = object (TIdentifier)
      Operands : array [0 .. MaxIdentifier] of Integer;
      IPriority, OPriority : array [0 .. MaxIdentifier] of Integer;
      procedure Register (S : TIdent; IP, OP, OPs : Integer);
    end;

  var
    Ress : TIdentifier; {Reserved Words}
    Vars : TVariable;   {Variables}
    Nums : TNumber;     {Constant Numbers}
    Oprs : TOperator;   {Operators}

  type
    TNode = record
      Ty : (Numb, Varb, Oper);
      N, M : Integer;
    end;

  type
    PExpression = ^ TExpression;
    TExpression = object (TObject)
      Nd : TNode;
      L, R : PExpression;
      constructor Init;
      procedure Parse;
      procedure Draw;
      procedure CodeGenerate;
      destructor Done; virtual;
    private
      ICol, MCol : Integer;
      function  Lable : TIdent;
      procedure WriteNode;
    end;

  type
    PStatement = ^ TStatement;
    TStatement = object (TObject)
      Next : PStatement;
      Expr : PExpression;
      constructor Init;
      procedure Parse;
      procedure Draw;
      procedure CodeGenerate;
      destructor Done; virtual;
    end;

  type
    PProgram = ^ TProgram;
    TProgram = object (TObject)
      Head : PStatement;
      constructor Init;
      procedure Parse;
      procedure Draw;
      procedure CodeGenerate;
      destructor Done; virtual;
    end;

implementation

var
  InAsm : Boolean;
  Depth, MaxDepth : Integer;
  Row, Col, MaxCol : Integer;
  Screen : array [0 .. 100] of string;

constructor TIdentifier.Init;
begin
  Size := 0;
end;

function  TIdentifier.Insert (S : TIdent) : Integer;
var
  I : Integer;

begin
  I := Index(S);
  if I = 0 then
  begin
    if I = MaxIdentifier then
      Error('Too Many Identifiers');
    Inc(Size);
    Name[Size] := UCase(S);
    Insert := Size;
  end
  else
   Insert := I;
end;

function  TIdentifier.Index (S : TIdent) : Integer;
var
  I : Integer;
begin
  for I := Size downto 0 do
    if Name[I] = UCase(S) then
      Break;
  Index := I;
end;

destructor TIdentifier.Done;
begin
  Init;
end;


function  TVariable.Insert (S : TIdent) : Integer;
begin
  if Ress.Index(S) <> 0 then
    CompileError('Identifier Expected - ' + S);
  Insert := inherited Insert(S);
end;

procedure TOperator.Register (S : TIdent; IP, OP, OPs : Integer);
var
  I : Integer;
begin
  I := Insert(S);
  IPriority[I] := IP;
  OPriority[I] := OP;
  Operands[I] := OPs;
end;


function EncodeNumber (S : TIdent) : TIdent;
begin
  while (S[0] > #1) and (S[Ord(S[0])] = '0') do Dec(S[0]);
  if S[Ord(S[0])] = '.' then Dec(S[0]);
  if S[1] = '-' then
    S[1] := '_';
  if Pos('.', S) <> 0 then
    S[Pos('.', S)] := 'o';
  S := 'N' + S;
  EncodeNumber := S;
end;

function DecodeNumber (S : TIdent) : TIdent;
begin
  Delete(S, 1, 1);
  if S[1] = '_' then
    S[1] := '-';
  if Pos('o', S) <> 0 then
    S[Pos('o', S)] := '.';
  if Pos('O', S) <> 0 then
    S[Pos('O', S)] := '.';
  DecodeNumber := S;
end;

function  TNumber.Insert (S : TIdent) : Integer;
begin
  Insert := inherited Insert(EncodeNumber(S));
end;

function  TNumber.Index (S : TIdent) : Integer;
begin
  Index := inherited Index(EncodeNumber(S));
end;

function  TNumber.InsertNum (X : Extended) : Integer;
var
  S : TIdent;
begin
  Str(X : 0 : 18, S);
  InsertNum := inherited Insert(EncodeNumber(S));
end;

function  TNumber.IndexNum (X : Extended) : Integer;
var
  S : TIdent;
begin
  Str(X : 0 : 18, S);
  IndexNum := inherited Index(EncodeNumber(S));
end;


constructor TExpression.Init;
begin
  Nd.Ty := Oper;
  L := nil;
  R := nil;
end;

procedure TExpression.Parse;
const
  MaxSize = MaxIdentifier;

type
  PStack = ^ TStack;
  TStack = array [1 .. MaxSize + 1] of TNode;
  TQueue = array [1 .. MaxSize + 1] of PExpression;

var
  Queue : TQueue;
  QSize : Integer;
  Stack : PStack;
  SSize : Integer;
  TN, T : TNode;
  I, J  : Integer;
  P     : Pointer;

  function  JoinTwoTrees (A, B : PExpression; N : TNode) : PExpression;
  var T : PExpression;
  begin
    New(T);
    T^.Nd := N;
	  T^.L := A;
	  T^.R := B;
    JoinTwoTrees := T;
  end;

  procedure Push;
  begin
    if SSize >= MaxSize then
      CompileError('Expression Too Complicated');
    Inc(SSize);
    Stack^[SSize] := T;
  end;

  procedure Pop;
  begin
    if SSize <= 0 then
      CompileError('Error In Expression');
    T := Stack^[SSize];
    Dec(SSize);
  end;

  procedure EnQueue;
  begin
    if QSize >= MaxSize then
      CompileError('Expression Too Complicated');
    Inc(QSize);
    Queue[QSize] := JoinTwoTrees(nil, nil, T);
  end;

  procedure DeQueue;
  var I : Integer;
  begin
    I := Oprs.Operands[T.N];
    if QSize < I then
      CompileError('Error In Expression');
    if I > 0 then
    begin
      Dec(QSize, I - 1);
      if I = 2 then
        Queue[QSize] := JoinTwoTrees(Queue[QSize], Queue[QSize + 1], T)
      else
        Queue[QSize] := JoinTwoTrees(nil,          Queue[QSize],     T)
    end;
  end;

  procedure ParseToken;
  begin
    if TN.Ty <> Oper then
    begin
      T := TN;
      Push;
      Exit;
    end;
    if (Oprs.Name[TN.N] = '=') and ((Stack^[SSize].Ty <> Varb) or
        ((Stack^[SSize - 1].Ty = Oper) and
         (Oprs.OPriority[TN.N] < Oprs.IPriority[Stack^[SSize - 1].N]))) then
      CompileError('LValue Expected.');
    if Stack^[SSize].Ty <> Oper then
    begin
      Pop;
      EnQueue;
    end;
    if Stack^[SSize].Ty <> Oper then
      Needs(';');
    with Oprs do
      while (SSize > 0) and (OPriority[TN.N] < IPriority[Stack^[SSize].N]) do
      begin Pop; DeQueue; end;
    if not ((Oprs.Name[TN.N] = ')') or (Oprs.Name[TN.N] = ';')) then
    begin
      T := TN;
      Push;
    end
    else
    begin
      Pop;
      if (SSize = 0) and (Oprs.Name[Tn.N] = ')') then
        Needs(';');
    end;
  end;

begin {Parse}
  New(Stack);
  SSize := 1;
  QSize := 0;
  with Stack^[1] do
  begin
    Ty := Oper;
    N := Oprs.Index('$');
  end;
  if Token = ';' then
    CompileError('Expression Expected');
  while True do
  begin
    with TN do
    begin
      case TokenType of
        Id:
        begin
          if Ress.Index(Token) = 0 then
          begin
            Ty := Varb;
            N := Vars.Insert(Token)
          end
          else
          if Ress.Index(Token) > 2 then
          begin
            Ty := Oper;
            N := Oprs.Index('"');
            M := Ress.Index(Token);
          end
          else
          begin
            Ty := Oper;
            N := Oprs.Index(';');
          end;
        end;
        Number: begin Ty := Numb; N := Nums.Insert(Token); end;
        Operator : begin Ty := Oper; N := Oprs.Index(Token); end;
        Null : begin Ty := Oper; N := Oprs.Index(';'); end;
      end;
      with Oprs do
        if (Ty = Oper) and (Name[N] = '-') and (Stack^[SSize].Ty = Oper) and
           not (Name[Stack^[SSize].N] = ')') then
          N := Index('~'); {check for neg (~)}
      ParseToken;
      if (Ty = Oper) and (Oprs.Name[N] = ';') then
        Break;
    end;
    NextToken;
  end;
  if (QSize <> 1) or (SSize <> 0) then
    CompileError('Error In Expression');
  L := Queue[1]^.L;
  R := Queue[1]^.R;
  Nd := Queue[1]^.Nd;
  Dispose(Queue[1]);
  Dispose(Stack);
end;

function  TExpression.Lable : TIdent;
var
  S : TIdent;
begin
  case Nd.Ty of
    Numb:
      S := DecodeNumber(Nums.Name[Nd.N]);
    Oper:
      if Oprs.Name[Nd.N] <> '"' then
        if Oprs.Name[Nd.N] = '~' then
          S := '-'
        else
          S := Oprs.Name[Nd.N]
      else
        S := Ress.Name[Nd.M];
    Varb:
      begin
        S := Vars.Name[Nd.N];
      end;
  end;
  Lable := S;
end;

procedure TExpression.WriteNode;
var
  S : TIdent;
  I, J : Integer;
begin
  S := Lable;
  J := MCol - Length(S) div 2;
  if J < 1 then
    J := 1;
  for I := 1 to Length(S) do
    Screen[Row, J + I - 1] := S[I];
  ICol := J + Length(S) - 1;
end;

procedure TExpression.Draw;
var
  P : PExpression;
  I, J, K : Integer;
begin
  Inc(Depth);
  if Depth > MaxDepth then
    MaxDepth := Depth;
  if L <> nil then P := L else if R <> nil then P := R else P := nil;
  if P <> nil then
  begin
    Inc(Row, 2);
    P^.Draw;
    Dec(Row, 2);
    ICol := P^.ICol;
    if P = R then
    begin
      MCol := P^.MCol;
      Screen[Row + 1, MCol] := CharLineDown;
      WriteNode;
      if P^.ICol > ICol then
        ICol := P^.ICol;
      Dec(Depth);
      Exit;
    end;
    J := L^.ICol + TreeSpace + 1;
    if J - Col - 3 < Length(Lable) then
      J := Col + 3 + Length(Lable);
    K := Col;
    Col := J;
    Inc(Row, 2);
    R^.Draw;
    Dec(Row, 2);
    Col := K;
    ICol := R^.ICol;
    MCol := (L^.MCol + R^.MCol) div 2;
    Inc(Row);
    for I := L^.MCol + 1 to R^.MCol - 1 do
      Screen[Row, I] := CharLineRight;
    Screen[Row, L^.MCol] := CharCornerUL;
    Screen[Row, R^.MCol] := CharCornerUR;
    Screen[Row, MCol] := CharTDown;
    Dec(Row);
    WriteNode;
    ICol := R^.ICol;
  end
  else
  begin
    MCol := (Col + Length(Lable) div 2);
    WriteNode;
  end;
  Dec(Depth);
end;

procedure TExpression.CodeGenerate;
begin
  if (Depth = 0) and InAsm then
    Writeln;
  Inc(Depth);
  if not InAsm then
  begin
    InAsm := True;
    Writeln('  asm');
  end;
  with Nd do
    case Ty of
      Varb: Writeln('    fld  _' + Vars.Name[N]);
      Numb: if DecodeNumber(Nums.Name[N]) = '0' then Writeln('    fld0') else
            if DecodeNumber(Nums.Name[N]) = '1' then Writeln('    fld1') else
            Writeln('    fld  ' + Nums.Name[N]);
      Oper:
      begin
        if (L <> nil) and not
           ((Nd.Ty = Oper) and (Oprs.Name[Nd.N] = '=')) then
          L^.CodeGenerate;
        R^.CodeGenerate;
        if not InAsm then
        begin
          InAsm := True;
          Writeln('  asm');
        end;
        if Oprs.Name[N] = '~' then Writeln('    fchs') else
        if Oprs.Name[N] = '+' then Writeln('    fadd') else
        if Oprs.Name[N] = '-' then Writeln('    fsub') else
        if Oprs.Name[N] = '*' then Writeln('    fmul') else
        if Oprs.Name[N] = '/' then Writeln('    fdiv') else
        if Oprs.Name[N] = '=' then
        begin
          if Depth > 1 then
            Writeln('    fld  st(0)');
          Writeln('    fstp _' + Vars.Name[L^.Nd.N]);
          Dec(Depth);
          Exit;
        end else
        if Oprs.Name[N] = '"' then
          case M of
            3: begin
                if Depth > 1 then
                  Writeln('    fld  st(0)');
                 Writeln('    fstp WriteVal');
                 Writeln('  end;');
                 InAsm := False;
                 Writeln('  Writeln(WriteVal:0:17);');
                 Dec(Depth);
                 Exit;
               end;
            4: Writeln('    fabs' );
            5: Writeln('    fsqrt');
            6: Writeln('    fsin' );
            7: Writeln('    fcos' );
            8: begin
               Writeln('    fld1'  );
               Writeln('    fxch'  );
               Writeln('    fyl2x' );
               Writeln('    fldl2e');
               Writeln('    fdiv'  );
               end;
          end;
      end;
    end;
  Dec(Depth);
  if Depth = 0 then
    Writeln('    fstp');
end;

destructor TExpression.Done;
begin
  Init;
end;


constructor TStatement.Init;
begin
  Next := nil;
  Expr := nil;
end;

procedure TStatement.Parse;
begin
  New(Expr, Init);
  Expr^.Parse;
  if Token = ';' then
  begin
    NextToken;
    if (Token = Ress.Name[2]) or (Token = '') then
      Exit;
    New(Next, Init);
    Next^.Parse;
  end;
end;

procedure TStatement.Draw;
var
  I, J : Integer;
begin
  FillChar(Screen, SizeOf(Screen), #32);
  MaxDepth := 0;
  Depth := 0;
  Row := 0;
  Col := 1;
  MaxCol := 0;
  if Expr <> nil then
    Expr^.Draw;
  for I := 0 to 2 * MaxDepth do
  begin
    for J := SizeOf(Screen[0]) - 1 downto 0 do
      if Screen[I, J] <> #32 then
        Break;
    Byte(Screen[I, 0]) := J;
    Writeln(Screen[I]);
  end;
  if Next <> nil then
    Next^.Draw;
end;

procedure TStatement.CodeGenerate;
begin
  Depth := 0;
  Expr^.CodeGenerate;
  if Next <> nil then
    Next^.CodeGenerate
  else
    if InAsm then
    begin
      Writeln('  end;');
      InAsm := False;
    end;
end;

destructor TStatement.Done;
begin
  if Next <> nil then
    Dispose(Next, Done);
  Init;
end;


constructor TProgram.Init;
begin
  Head := nil;
  Vars.Init;
  with Nums do
  begin
    Init;
    InsertNum(0);
    InsertNum(1);
  end;
  with Ress do
  begin
    Init;
    Insert('begin');
    Insert('end');
    Insert('write');
    Insert('abs');
    Insert('sqrt');
    Insert('sin');
    Insert('cos');
    Insert('ln');
  end;
  with Oprs do
  begin
    Init;
    Register('$', 0, 20, 0); {begin of expression/stack}
    Register('(', 0, 20, 0);
    Register('+', 4, 3, 2);
    Register('-', 4, 3, 2);
    Register('*', 6, 5, 2);
    Register('/', 6, 5, 2);
    Register('~', 7, 8, 1); {neg operator}
    Register(';', 0, 0, 0); {end of expression/stack}
    Register(')', 0, 0, 0);
    Register('=', 1, 2, 2); {assignment operator}
    Register('"', 7, 8, 1); {function call operator}
  end;
end;

procedure TProgram.Parse;
begin
  if Token = Ress.Name[1] then
    ExpireToken;
  if Token <> Ress.Name[2] then
  begin
    New(Head, Init);
    Head^.Parse;
  end;
  if Token = Ress.Name[2] then
    ExpireToken
  else
    if not SeekEOF then
      CompileError('; or end expected.');
end;

procedure TProgram.Draw;
begin
  Writeln;
  if Head <> nil then
    Head^.Draw;
  Writeln
end;

procedure TProgram.CodeGenerate;
var
  I : Integer;
begin
  InAsm := False;
  Writeln('{Generated by Expression Parser  version 1.0'#13#10 +
          ' by Behdad Esfahbod  Feb 2001}');
  Writeln('{$g+,n+,e-}'#13#10'{$m 1024,0,0}');
  if Nums.Size > 2 then
  begin
    Writeln('const');
    for I := 3 to Nums.Size do
      Writeln('  ' + Nums.Name[I] + ' : Extended = ' +
                     DecodeNumber(Nums.Name[I]) + ';');
  end;
  Writeln('var');
  Writeln('  WriteVal : Extended;');
  if Vars.Size > 0 then
  begin
    for I := 1 to Vars.Size do
      Writeln('  _' + Vars.Name[I] + ' : Extended;');
  end;
  Writeln('begin');
  if Vars.Index('pi') <> 0 then
  begin
    InAsm := True;
    Writeln('  asm'#13#10 +
            '    fldpi'#13#10 +
            '    fstp _' + Vars.Name[Vars.Index('pi')]);
  end;
  if Head <> nil then
    Head^.CodeGenerate;
  if InAsm then
    Writeln('  end;');
  Writeln('end.');
end;

destructor TProgram.Done;
begin
  if Head <> nil then
    Dispose(Head, Done);
  Init;
end;

end.
