{$A+,B-,D+,E-,F-,G+,I-,L+,N+,O-,P-,Q-,R-,S-,T-,V+,X+,Y+}
{by Behdad Esfahbod}
{Define Scaner, Vars, Consts, Types and Needed Funcs and Procs}
unit
  _Global;

interface

  const
    TreeSpace     = 3;
    CharLineDown  = '';
    CharLineRight = '';
    CharCornerUR  = '';
    CharCornerUL  = '';
    CharTLeft     = '';
    CharTDown     = '';

  const
    MaxIdentifier = 200;
    IdentifierLen = 20;

  type
    TTokenType = (Null, Id, Number, Operator);
    TIdent = string [IdentifierLen];

  const
    Letter = ['A' .. 'Z', '_'];
    Digit  = ['0' .. '9'];
    Operat = [';', '(', ')', '+', '-', '*', '/', '='];
    Other  = ['.'];
    ValidChars = Letter + Digit + Operat + Other;

  procedure Error (Msg : string);
  procedure CompileError (Msg : string);

  procedure InitIO;
  procedure DoneIO;

  function  TokenType : TTokenType;
  function  NextTokenType : TTokenType;
  function  Token : string;
  function  NextToken : string;
  procedure ExpireToken;

  procedure Needs (S : string);

  function  UCase (S : string) : string;
  function  I2S (A : Integer) : string;

implementation

uses
  _Expression;

procedure Header;
begin
  Writeln('Expression Parser  by Behdad Esfahbod  Feb 2001'#13#10);
end;

procedure Help;
begin
  Writeln('Usage :'#13#10#32#32'exprtree input[.exp]'#13#10#13#10);
  Halt;
end;

const
  Row : Integer = 1;
  Col : Integer = 0;

procedure CompileError (Msg : string);
begin
  Assign(Output, '');
  Rewrite(Output);
  Write('Compile Error at Line ', Row, { ', Column ', Col,} ': ' +
        #13#10#32#32 + Msg + #13#10#7);
  Close(Output);
  Halt(2);
end;

procedure Error (Msg : string);
begin
  Assign(Output, '');
  Rewrite(Output);
  Write('Error : '#13#10#32#32 + Msg + #13#10#7);
  Close(Output);
  Halt(1);
end;

var
  InBuffer, OutBuffer : array [0 .. 4095] of Byte;
  S1, S2 : string;

procedure InitIO;
begin
  Header;
  {$I-}
  if ParamCount <> 1 then
    Help;
  S1 := ParamStr(1);
  if Pos('.', S1) = 0 then
    S1 := S1 + '.exp';
  Assign(Input, S1);
  Reset(Input);
  if IOResult <> 0 then
    Error('Opening Input File - ' + S1);
  SetTextBuf(Input, InBuffer);
  S2 := S1;
  if Pos('.', S2) > 0 then
    S2 := Copy(S2, 1, Pos('.', S2) - 1) + '.pas';
  Assign(Output, S2);
  Rewrite(Output);
  if IOResult <> 0 then
    Error('Opening Output File - ' + S2);
  SetTextBuf(Output, OutBuffer);
end;

procedure DoneIO;
begin
  Close(Input);
  Close(Output);
  Assign(Output, '');
  Rewrite(Output);
  Writeln(S1 + ' Compiled Successfully.'#13#10'Output Written to ' +
          S2 + #13#10);
end;

const
  NewLine : Boolean = False;
  TheTokenType : TTokenType = Null;
  TheToken : string = '';
  ValidToken : Boolean = False;
  ValidChar : Boolean = False;
  TheChar : Char = #0;

function  TokenType : TTokenType;
begin
  Token;
  TokenType := TheTokenType;
end;

function  NextTokenType : TTokenType;
begin
  NextToken;
  NextTokenType := TheTokenType;
end;

procedure BufferChar (C : Char);
begin
  ValidChar := True;
  TheChar := C;
end;

function  ReadRealChar : Char;
var
  C : Char;

begin
  C := #0;
  if not Eof then
  begin
    Read(C);
    C := UpCase(C);
    if C = #10 then
      NewLine := True
    else
    begin
      if NewLine then
      begin
        Inc(Row);
        Col := 0;
      end;
      NewLine := False;
    end;
    if C = #8 then
      Inc(Col, 2)
    else
      Inc(Col);
  end;
  ReadRealChar := C;
end;

function  ReadChar : Char;
var
  C : Char;

begin
  if ValidChar then
  begin
    ValidChar := False;
    ReadChar := TheChar;
    Exit;
  end;
  C := #0;
  while not Eof and not (C in ValidChars) do
    C := ReadRealChar;
  ReadChar := C;
end;

procedure ReadToken;
var
  S : string;
  C : Char;

begin
  S := '';
  TheToken := S;
  TheTokenType := Null;
  if ValidChar or not SeekEof then
  begin
    C := ReadChar;
    if C in Letter then
      TheTokenType := Id
    else
    if C in Digit then
      TheTokenType := Number
    else
    if C in Operat then
      TheTokenType := Operator;
  end
  else
    Exit;
  S := C;
  C := ReadRealChar;
  case TheTokenType of
    Id:
      while C in Letter + Digit do
      begin
        S := S + C;
        C := ReadRealChar;
      end;
    Number:
    begin
      while C in Digit do
      begin
        S := S + C;
        C := ReadRealChar;
      end;
      if C = '.' then
      begin
        S := S + C;
        C := ReadRealChar;
        while C in Digit do
        begin
          S := S + C;
          C := ReadRealChar;
        end;
      end;
    end;
    Operator:
    begin
      while Oprs.Index(S) = 0 do
      begin
        S := S + C;
        C := ReadRealChar;
        if not (C in Operat) then
          Break;
      end;
      if Oprs.Index(S) = 0 then
        CompileError('Invalid Operator - ' + S);
    end;
    Null:
      S := '';
  end;
  if C in ValidChars then
    BufferChar(C);
  TheToken := S;
end;

function  Token : string;
begin
  if not ValidToken then
  begin
    ValidToken := True;
    ReadToken;
  end;
  Token := TheToken
end;

function  NextToken : string;
begin
  Token;
  ValidToken := False;
  NextToken := Token;
end;

procedure ExpireToken;
begin
  ValidToken := False;
end;

function  UCase (S : string) : string;
var
  I : Integer;

begin
  for I := 1 to Length(S) do
    S[I] := UpCase(S[I]);
  UCase := S;
end;

function  I2S (A : Integer) : string;
var
  S : string;

begin
  Str(A, S);
  I2S := S;
end;

procedure Needs (S : string);
begin
  if Token <> UCase(S) then
    CompileError(S + ' Expected');
  NextToken;
end;

end.
